@@ -4,11 +4,12 @@ import abc
 import enum
 import json
 import math
+import types
 
 try:
     from collections.abc import MutableSequence  # only works on python 3.3+
 except ImportError:  # pragma: no cover
-    from collections import MutableSequence
+    from collections import MutableSequence  # pylint: disable=deprecated-class
 
 from collections import OrderedDict
 
@@ -184,7 +185,7 @@ class Serializable(object):  # pylint: disable=too-few-public-methods
             return False, obj.name
         elif hasattr(obj, '__dict__') or isinstance(obj, dict):
             result = cls._markdown_result_complex(obj, level)
-        elif isinstance(obj, (list, tuple, set)):
+        elif isinstance(obj, (list, tuple, set, ArrayBase)):
             result = cls._markdown_result_list(obj, level)
         else:
             result = False, str(obj)
@@ -304,8 +305,11 @@ class OpaqueParam(VectorParamNumeric):  # pylint: disable=too-few-public-methods
 class VectorParamString(VectorParamBase):  # pylint: disable=too-few-public-methods
     separator = attr.ib(validator=attr.validators.instance_of(six.string_types), default=',')
     encoding = attr.ib(validator=attr.validators.instance_of(six.string_types), default='ascii')
-    item_class = attr.ib(validator=attr.validators.instance_of(type), default=str)
-    fallback_class = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(type)), default=None)
+    item_class = attr.ib(validator=attr.validators.instance_of((type, types.FunctionType)), default=str)
+    fallback_class = attr.ib(
+        default=None,
+        validator=attr.validators.optional(attr.validators.instance_of((type, types.FunctionType)))
+    )
 
     def get_item_size(self, item):
         if isinstance(item, (ParsableBase, StringEnumParsable)):
@@ -318,8 +322,10 @@ class VectorParamString(VectorParamBase):  # pylint: disable=too-few-public-meth
 
 @attr.s
 class VectorParamParsable(VectorParamBase):  # pylint: disable=too-few-public-methods
-    item_class = attr.ib(validator=attr.validators.instance_of(type))
-    fallback_class = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(type)))
+    item_class = attr.ib(validator=attr.validators.instance_of((type, types.FunctionType)))
+    fallback_class = attr.ib(
+        validator=attr.validators.optional(attr.validators.instance_of((type, types.FunctionType)))
+    )
 
     def get_item_size(self, item):
         return len(item.compose())
@@ -391,8 +397,11 @@ class ArrayBase(ParsableBase, MutableSequence, Serializable):
     def append(self, value):
         self.insert(len(self._items), value)
 
+    def _asdict(self):
+        return self._items
+
     def _as_markdown(self, level):
-        return self._markdown_result(self._items, level)
+        return self._markdown_result(self._asdict(), level)
 
 
 class Vector(ArrayBase):
@@ -526,13 +535,12 @@ class VectorParsableDerived(ArrayBase):
         return header_composer.composed_bytes + body_composer.composed_bytes
 
 
-@attr.s(init=False)
 class Opaque(ArrayBase):
-    def __init__(self, items):
-        if isinstance(items, (bytes, bytearray)):
-            items = [ord(items[i:i + 1]) for i in range(len(items))]
+    def __attrs_post_init__(self):
+        if isinstance(self._items, (bytes, bytearray)):
+            self._items = [ord(self._items[i:i + 1]) for i in range(len(self._items))]
 
-        super(Opaque, self).__init__(items)
+        super(Opaque, self).__attrs_post_init__()
 
     @classmethod
     def _parse(cls, parsable):
@@ -769,3 +777,47 @@ class ListParsable(ArrayBase):
             composer.compose_raw(separator)
 
         return composer.composed_bytes
+
+
+class OpaqueEnumParsable(Vector):
+    @classmethod
+    def _parse(cls, parsable):
+        opaque, parsed_length = super(OpaqueEnumParsable, cls)._parse(parsable)
+        code = bytearray(opaque).decode(cls.get_encoding())
+
+        try:
+            parsed_object = next(iter([
+                enum_item
+                for enum_item in cls.get_enum_class()
+                if enum_item.value.code == code
+            ]))
+        except StopIteration as e:
+            six.raise_from(InvalidValue(code.encode(cls.get_encoding()), cls), e)
+
+        return parsed_object, parsed_length
+
+    @classmethod
+    @abc.abstractmethod
+    def get_enum_class(cls):
+        raise NotImplementedError()
+
+    @classmethod
+    def get_encoding(cls):
+        return 'utf-8'
+
+
+class OpaqueEnumComposer(enum.Enum):
+    def __repr__(self):
+        return self.__class__.__name__ + '.' + self.name
+
+    def compose(self):
+        composer = ComposerBinary()
+        value = self.value.code.encode(self.get_encoding())  # pylint: disable=no-member
+
+        composer.compose_bytes(value, 1)
+
+        return composer.composed_bytes
+
+    @classmethod
+    def get_encoding(cls):
+        return 'utf-8'
