@@ -1,13 +1,12 @@
-#!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
 import abc
 import collections
 import enum
 import random
-import six
 
 import attr
+import six
 
 from cryptoparser.common.base import (
     VariantParsable,
@@ -25,7 +24,13 @@ from cryptoparser.ssh.ciphersuite import (
     SshMacAlgorithm,
     SshCompressionAlgorithm,
 )
-from cryptoparser.ssh.version import SshProtocolVersion
+from cryptoparser.ssh.key import SshPublicKeyBase, SshHostPublicKeyVariant
+from cryptoparser.ssh.version import (
+    SshProtocolVersion,
+    SshSoftwareVersionBase,
+    SshSoftwareVersionParsedVariant,
+    SshSoftwareVersionUnparsed,
+)
 
 
 class SshMessageCode(enum.IntEnum):
@@ -74,18 +79,9 @@ class SshMessageBase(ParsableBase):
 @attr.s
 class SshProtocolMessage(ParsableBase):
     protocol_version = attr.ib(validator=attr.validators.instance_of(SshProtocolVersion))
-    software_version = attr.ib(validator=attr.validators.instance_of(six.string_types))
+    software_version = attr.ib(validator=attr.validators.instance_of(SshSoftwareVersionBase))
     comment = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(six.string_types)), default=None)
 
-    @software_version.validator
-    def software_version_validator(self, _, value):  # pylint: disable=no-self-use
-        if '\r' in value or '\n' in value or ' ' in value:
-            raise InvalidValue(value, SshProtocolMessage, 'software_version')
-        try:
-            value.encode('ascii')
-        except UnicodeEncodeError as e:
-            six.raise_from(InvalidValue(value, SshProtocolMessage, 'software_version'), e)
-
     @comment.validator
     def comment_validator(self, _, value):  # pylint: disable=no-self-use
         if value is not None:
@@ -114,7 +110,12 @@ class SshProtocolMessage(ParsableBase):
         if software_version_and_comment[-1][-1] == '\r':
             software_version_and_comment[-1] = software_version_and_comment[-1][:-1]
 
-        software_version = software_version_and_comment[0]
+        software_version_parser = ParserText(software_version_and_comment[0].encode('ascii'))
+        try:
+            software_version_parser.parse_parsable('value', SshSoftwareVersionParsedVariant)
+        except InvalidValue:
+            software_version_parser.parse_parsable('value', SshSoftwareVersionUnparsed)
+
         if len(software_version_and_comment) > 1:
             comment = ' '.join(software_version_and_comment[1:])
         else:
@@ -124,7 +125,11 @@ class SshProtocolMessage(ParsableBase):
         if parser.parsed_length > 255:
             raise TooMuchData(parser.parsed_length - 255)
 
-        return SshProtocolMessage(parser['protocol_version'], software_version, comment), parser.parsed_length
+        return SshProtocolMessage(
+            parser['protocol_version'],
+            software_version_parser['value'],
+            comment
+        ), parser.parsed_length
 
     def compose(self):
         composer = ComposerText()
@@ -419,7 +424,7 @@ class SshDHGroupExchangeInit(SshDHKeyExchangeInitBase):
 
 @attr.s
 class SshDHKeyExchangeReplyBase(SshMessageBase):
-    host_public_key = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
+    host_public_key = attr.ib(validator=attr.validators.instance_of(SshPublicKeyBase))
     ephemeral_public_key = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
     signature = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
 
@@ -430,7 +435,7 @@ class SshDHKeyExchangeReplyBase(SshMessageBase):
 
     @staticmethod
     def _parse_ssh_key(parser):
-        parser.parse_bytes('host_public_key', 4)
+        parser.parse_parsable('host_public_key', SshHostPublicKeyVariant, 4)
         parser.parse_bytes('ephemeral_public_key', 4)
         parser.parse_bytes('signature', 4)
 
@@ -438,7 +443,7 @@ class SshDHKeyExchangeReplyBase(SshMessageBase):
 
     @staticmethod
     def _compose_ssh_key(composer, host_public_key, ephemeral_public_key_bytes, signature_bytes):
-        composer.compose_bytes(host_public_key, 4)
+        composer.compose_parsable(host_public_key, 4)
         composer.compose_bytes(ephemeral_public_key_bytes, 4)
         composer.compose_bytes(signature_bytes, 4)
 
