@@ -7,10 +7,12 @@ import enum
 import six
 import attr
 
+from cryptoparser.tls.algorithm import TlsNextProtocolName, TlsProtocolName
 from cryptoparser.common.base import (
     Opaque,
     OpaqueParam,
     TwoByteEnumComposer,
+    OpaqueEnumParsable,
     TwoByteEnumParsable,
     VariantParsable,
     Vector,
@@ -19,7 +21,7 @@ from cryptoparser.common.base import (
     VectorParsable,
     VectorParsableDerived,
 )
-from cryptoparser.common.exception import NotEnoughData, InvalidType
+from cryptoparser.common.exception import NotEnoughData, InvalidType, InvalidValue
 from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
 from cryptoparser.tls.algorithm import (
     TlsNamedCurve,
@@ -301,7 +303,7 @@ class TlsExtensionUnparsed(TlsExtensionBase):
 
     @classmethod
     def _parse(cls, parsable):
-        parser = super(TlsExtensionUnparsed, cls)._check_header(parsable)
+        parser = cls._check_header(parsable)
 
         parser.parse_raw('extension_data', parser['extension_length'])
 
@@ -311,12 +313,7 @@ class TlsExtensionUnparsed(TlsExtensionBase):
         composer.compose_parsable(self.extension_type)
 
     def compose(self):
-        payload_composer = ComposerBinary()
-        payload_composer.compose_raw(self.extension_data)
-
-        header_bytes = self._compose_header(payload_composer.composed_length)
-
-        return header_bytes + payload_composer.composed_bytes
+        return self._compose_header(len(self.extension_data)) + self.extension_data
 
 
 @attr.s
@@ -348,6 +345,27 @@ class TlsExtensionParsed(TlsExtensionBase):
         return parser
 
 
+class TlsExtensionUnusedData(TlsExtensionParsed):
+    @classmethod
+    def _parse(cls, parsable):
+        parser = cls._parse_header(parsable)
+
+        parser.parse_raw('extension_data', parser['extension_length'])
+
+        if parser['extension_data']:
+            raise InvalidValue(parser['extension_data'], cls)
+
+        return cls(), parser.parsed_length
+
+    def compose(self):
+        return self._compose_header(0)
+
+    @classmethod
+    @abc.abstractmethod
+    def get_extension_type(cls):
+        raise NotImplementedError()
+
+
 class TlsServerNameType(enum.IntEnum):
     HOST_NAME = 0x00
 
@@ -823,6 +841,197 @@ class TlsExtensionCertificateStatusRequest(TlsExtensionParsed):
         return header_bytes + payload_composer.composed_bytes
 
 
+class TlsRenegotiatedConnection(Opaque):
+    @classmethod
+    def get_param(cls):
+        return OpaqueParam(
+            min_byte_num=0,
+            max_byte_num=2 ** 8 - 1,
+        )
+
+
+@attr.s
+class TlsExtensionRenegotiationInfo(TlsExtensionParsed):
+    renegotiated_connection = attr.ib(
+        default=TlsRenegotiatedConnection([]),
+        validator=attr.validators.instance_of(TlsRenegotiatedConnection)
+    )
+
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.RENEGOTIATION_INFO
+
+    @classmethod
+    def _parse(cls, parsable):
+        parser = super(TlsExtensionRenegotiationInfo, cls)._parse_header(parsable)
+
+        parser.parse_parsable('renegotiated_connection', TlsRenegotiatedConnection)
+
+        return TlsExtensionRenegotiationInfo(parser['renegotiated_connection']), parser.parsed_length
+
+    def compose(self):
+        payload_composer = ComposerBinary()
+
+        payload_composer.compose_parsable(self.renegotiated_connection)
+
+        header_bytes = self._compose_header(payload_composer.composed_length)
+
+        return header_bytes + payload_composer.composed_bytes
+
+
+@attr.s
+class TlsExtensionSessionTicket(TlsExtensionParsed):
+    session_ticket = attr.ib(
+        default=bytearray([]),
+        validator=attr.validators.instance_of((bytes, bytearray))
+    )
+
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.SESSION_TICKET
+
+    @classmethod
+    def _parse(cls, parsable):
+        parser = super(TlsExtensionSessionTicket, cls)._parse_header(parsable)
+
+        parser.parse_raw('session_ticket', parser['extension_length'])
+
+        return TlsExtensionSessionTicket(parser['session_ticket']), parser.parsed_length
+
+    def compose(self):
+        payload_composer = ComposerBinary()
+
+        payload_composer.compose_raw(self.session_ticket)
+
+        header_bytes = self._compose_header(payload_composer.composed_length)
+
+        return header_bytes + payload_composer.composed_bytes
+
+
+class TlsProtocolNameFactory(OpaqueEnumParsable):
+    @classmethod
+    def get_enum_class(cls):
+        return TlsProtocolName
+
+    @classmethod
+    def get_param(cls):
+        return OpaqueParam(
+            min_byte_num=1, max_byte_num=2 ** 8 - 1
+        )
+
+
+class TlsProtocolNameList(VectorParsable):
+    @classmethod
+    def get_param(cls):
+        return VectorParamParsable(
+            item_class=TlsProtocolNameFactory,
+            fallback_class=None,
+            min_byte_num=2, max_byte_num=2 ** 16 - 1
+        )
+
+
+@attr.s
+class TlsExtensionApplicationLayerProtocolNegotiation(TlsExtensionParsed):
+    protocol_names = attr.ib(
+        converter=TlsProtocolNameList,
+        validator=attr.validators.instance_of(TlsProtocolNameList),
+    )
+
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.APPLICATION_LAYER_PROTOCOL_NEGOTIATION
+
+    @classmethod
+    def _parse(cls, parsable):
+        parser = super(TlsExtensionApplicationLayerProtocolNegotiation, cls)._parse_header(parsable)
+
+        parser.parse_parsable('protocol_names', TlsProtocolNameList)
+
+        return TlsExtensionApplicationLayerProtocolNegotiation(parser['protocol_names']), parser.parsed_length
+
+    def compose(self):
+        payload_composer = ComposerBinary()
+
+        payload_composer.compose_parsable(self.protocol_names)
+
+        header_bytes = self._compose_header(payload_composer.composed_length)
+
+        return header_bytes + payload_composer.composed_bytes
+
+
+class TlsExtensionNextProtocolNegotiationClient(TlsExtensionUnusedData):
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION
+
+
+class TlsNextProtocolNameFactory(OpaqueEnumParsable):
+    @classmethod
+    def get_enum_class(cls):
+        return TlsNextProtocolName
+
+    @classmethod
+    def get_param(cls):
+        return OpaqueParam(
+            min_byte_num=1, max_byte_num=2 ** 8 - 1
+        )
+
+
+class TlsNextProtocolNameList(VectorParsable):
+    @classmethod
+    def get_param(cls):
+        return VectorParamParsable(
+            item_class=TlsNextProtocolNameFactory,
+            fallback_class=None,
+            min_byte_num=1, max_byte_num=2 ** 16 - 1
+        )
+
+
+@attr.s
+class TlsExtensionNextProtocolNegotiationServer(TlsExtensionParsed):
+    protocol_names = attr.ib(
+        converter=TlsNextProtocolNameList,
+        validator=attr.validators.instance_of(TlsNextProtocolNameList),
+    )
+
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION
+
+    @classmethod
+    def _parse(cls, parsable):
+        parser = ParserBinary(parsable)
+
+        cls._parse_type(parser, 'extension_type')
+        if parser['extension_type'] != cls.get_extension_type():
+            raise InvalidType()
+
+        parser.parse_parsable('protocol_names', TlsNextProtocolNameList)
+
+        return cls(parser['protocol_names']), parser.parsed_length
+
+    def compose(self):
+        payload_composer = ComposerBinary()
+        payload_composer.compose_parsable(self.protocol_names)
+
+        header_composer = ComposerBinary()
+        self._compose_type(header_composer)
+
+        return header_composer.composed_bytes + payload_composer.composed_bytes
+
+
+class TlsExtensionEncryptThenMAC(TlsExtensionUnusedData):
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.ENCRYPT_THEN_MAC
+
+
+class TlsExtensionExtendedMasterSecret(TlsExtensionUnusedData):
+    @classmethod
+    def get_extension_type(cls):
+        return TlsExtensionType.EXTENDED_MASTER_SECRET
+
+
 class TlsExtensionVariantBase(VariantParsable):
     @classmethod
     @abc.abstractmethod
@@ -846,7 +1055,14 @@ class TlsExtensionVariantClient(TlsExtensionVariantBase):
     @classmethod
     def _get_parsed_extensions(cls):
         return collections.OrderedDict([
+            (TlsExtensionType.APPLICATION_LAYER_PROTOCOL_NEGOTIATION,
+                [TlsExtensionApplicationLayerProtocolNegotiation, ]),
+            (TlsExtensionType.ENCRYPT_THEN_MAC, [TlsExtensionEncryptThenMAC, ]),
+            (TlsExtensionType.EXTENDED_MASTER_SECRET, [TlsExtensionExtendedMasterSecret, ]),
+            (TlsExtensionType.RENEGOTIATION_INFO, [TlsExtensionRenegotiationInfo, ]),
+            (TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION, [TlsExtensionNextProtocolNegotiationClient, ]),
             (TlsExtensionType.SERVER_NAME, [TlsExtensionServerName, ]),
+            (TlsExtensionType.SESSION_TICKET, [TlsExtensionSessionTicket, ]),
             (TlsExtensionType.SUPPORTED_GROUPS, [TlsExtensionEllipticCurves, ]),
             (TlsExtensionType.EC_POINT_FORMATS, [TlsExtensionECPointFormats, ]),
             (TlsExtensionType.KEY_SHARE, [TlsExtensionKeyShareClient, ]),
@@ -859,6 +1075,14 @@ class TlsExtensionVariantServer(TlsExtensionVariantBase):
     @classmethod
     def _get_parsed_extensions(cls):
         return collections.OrderedDict([
+            (TlsExtensionType.APPLICATION_LAYER_PROTOCOL_NEGOTIATION,
+                [TlsExtensionApplicationLayerProtocolNegotiation, ]),
+            (TlsExtensionType.EC_POINT_FORMATS, [TlsExtensionECPointFormats, ]),
+            (TlsExtensionType.ENCRYPT_THEN_MAC, [TlsExtensionEncryptThenMAC, ]),
+            (TlsExtensionType.EXTENDED_MASTER_SECRET, [TlsExtensionExtendedMasterSecret, ]),
             (TlsExtensionType.KEY_SHARE, [TlsExtensionKeyShareClientHelloRetry, TlsExtensionKeyShareServer]),
+            (TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION, [TlsExtensionNextProtocolNegotiationServer, ]),
+            (TlsExtensionType.RENEGOTIATION_INFO, [TlsExtensionRenegotiationInfo, ]),
+            (TlsExtensionType.SESSION_TICKET, [TlsExtensionSessionTicket, ]),
             (TlsExtensionType.SUPPORTED_VERSIONS, [TlsExtensionSupportedVersionsServer, ]),
         ])
